package fitconnect.impl.crypto;

import com.nimbusds.jose.*;
import com.nimbusds.jose.crypto.RSADecrypter;
import com.nimbusds.jose.crypto.RSAEncrypter;
import com.nimbusds.jose.jwk.RSAKey;
import fitconnect.api.crypto.JWECryptoService;
import fitconnect.api.problems.DecryptionProblem;
import fitconnect.api.problems.EncryptionProblem;

import javax.crypto.KeyGenerator;
import javax.crypto.SecretKey;
import java.security.NoSuchAlgorithmException;
import java.security.interfaces.RSAPublicKey;
import java.text.ParseException;

public class FitCoCryptoService implements JWECryptoService {

    public static final JWEAlgorithm ALGORITHM = JWEAlgorithm.RSA_OAEP_256;
    public static final EncryptionMethod ENCRYPTION_METHOD = EncryptionMethod.A256GCM;

    @Override
    public String encryptString(RSAKey publicKey, String data) throws EncryptionProblem {
        final Payload payload = getPayloadString(data);
        return encrypt(publicKey, payload);
    }

    @Override
    public String decryptString(RSAKey privateKey, String encryptedData) throws DecryptionProblem {
        return decrypt(privateKey, encryptedData).toString();
    }

    @Override
    public String encryptBytes(RSAKey publicKey, byte[] bytes) throws EncryptionProblem {
        final Payload payload = getPayloadBytes(bytes);
        return encrypt(publicKey, payload);
    }

    @Override
    public byte[] decryptBytes(RSAKey privateKey, String encryptedData) throws DecryptionProblem {
        return decrypt(privateKey, encryptedData).toBytes();
    }

    private String encrypt(RSAKey publicKey, Payload payload) throws EncryptionProblem {
        try {
            KeyGenerator keyGenerator = KeyGenerator.getInstance("AES");
            keyGenerator.init(ENCRYPTION_METHOD.cekBitLength());
            final SecretKey cek = keyGenerator.generateKey();
            final String keyID = getIdFromPublicKey(publicKey);
            return encryptPayload(publicKey,payload, cek, keyID);
        } catch (NoSuchAlgorithmException | JOSEException e) {
            throw new EncryptionProblem(e.getMessage());
        }
    }

    private Payload decrypt(RSAKey privateKey, String encData) throws DecryptionProblem {
        try {
            final JWEObject jwe = JWEObject.parse(encData);
            jwe.decrypt(new RSADecrypter(privateKey));
            return jwe.getPayload();
        } catch (ParseException | JOSEException e) {
            throw new DecryptionProblem(e.getMessage());
        }
    }

    private JWEObject getJWEObject(String keyID, Payload payload) {
        return new JWEObject(getJWEHeader(keyID), payload);
    }

    private JWEHeader getJWEHeader(String keyID) {
        return new JWEHeader.Builder(ALGORITHM, ENCRYPTION_METHOD)
                .compressionAlgorithm(CompressionAlgorithm.DEF)
                .contentType("application/json")
                .keyID(keyID)
                .build();
    }

    private RSAEncrypter getEncrypter(RSAPublicKey publicKey, SecretKey cek) {
        return new RSAEncrypter(publicKey, cek);
    }

    private Payload getPayloadString(String data) {
        return new Payload(data);
    }

    private Payload getPayloadBytes(byte[] bytes) {
        return new Payload(bytes);
    }

    private String encryptPayload(RSAKey publicKey, Payload payload, SecretKey cek, String keyID) throws JOSEException {
        JWEObject jwe = getJWEObject(keyID, payload);
        jwe.encrypt(getEncrypter(publicKey.toRSAPublicKey(), cek));
        checkIfJWEObjectIsEncrypted(jwe);
        return jwe.serialize();
    }

    private void checkIfJWEObjectIsEncrypted(JWEObject jwe) {
        if (!jwe.getState().equals(JWEObject.State.ENCRYPTED)) {
            throw new EncryptionProblem("JWE object is not encrypted");
        }
    }

    private String getIdFromPublicKey(RSAKey publicKey) {
        final String keyID = publicKey.getKeyID();
        if (keyID == null || keyID.isEmpty()) {
            throw new EncryptionProblem("public key has no keyID");
        }
        return keyID;
    }
}
